#include <catch2/catch_test_macros.hpp>
#include <catch2/matchers/catch_matchers_all.hpp>

#include <algebra/Givens.hpp>
#include <algebra/SmallMatrix.hpp>
#include <algebra/SmallVector.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("Givens", "[algebra]")
{
  SECTION("classic (single vector)")
  {
    SECTION("square matrix")
    {
      SmallMatrix<double> A{5, 5};
      A.fill(0);
      A(0, 0) = 2;
      A(0, 1) = -1;

      A(1, 0) = -0.2;
      A(1, 1) = 2;
      A(1, 2) = -1;

      A(2, 1) = -1;
      A(2, 2) = 4;
      A(2, 3) = -2;

      A(3, 2) = -1;
      A(3, 3) = 2;
      A(3, 4) = -0.1;

      A(4, 3) = 1;
      A(4, 4) = 3;

      SmallVector<const double> x_exact = [] {
        SmallVector<double> y{5};
        y[0] = 1;
        y[1] = 3;
        y[2] = 2;
        y[3] = 4;
        y[4] = 5;
        return y;
      }();

      SmallVector<double> b = A * x_exact;

      SmallVector<double> x{5};
      x = zero;

      Givens::solve(A, x, b);
      SmallVector<double> error = x - x_exact;

      REQUIRE(std::sqrt(dot(error, error)) < 1E-10 * std::sqrt(dot(x, x)));
    }

    SECTION("rectangular matrix")
    {
      SmallMatrix<double> A{5, 3};
      A.fill(0);
      A(0, 0) = 2;
      A(0, 1) = -1;

      A(1, 0) = -0.2;
      A(1, 1) = 2;
      A(1, 2) = -1;

      A(2, 1) = -1;
      A(2, 2) = 4;

      A(3, 2) = -1;
      A(4, 0) = 0.5;
      A(4, 1) = 1;
      A(4, 2) = 1;

      SmallVector<const double> x_exact = [] {
        SmallVector<double> y{3};
        y[0] = 1;
        y[1] = 3;
        y[2] = 2;
        return y;
      }();

      SmallVector<double> b = A * x_exact;

      SmallVector<double> x{3};
      x = zero;

      Givens::solve(A, x, b);
      SmallVector<double> error = x - x_exact;

      REQUIRE(std::sqrt(dot(error, error)) < 1E-10 * std::sqrt(dot(x, x)));
    }
  }

  SECTION("generalized (vector collection)")
  {
    SECTION("square matrix")
    {
      SmallMatrix<double> A{5, 5};
      A.fill(0);
      A(0, 0) = 2;
      A(0, 1) = -1;

      A(1, 0) = -0.2;
      A(1, 1) = 2;
      A(1, 2) = -1;

      A(2, 1) = -1;
      A(2, 2) = 4;
      A(2, 3) = -2;

      A(3, 2) = -1;
      A(3, 3) = 2;
      A(3, 4) = -0.1;

      A(4, 3) = 1;
      A(4, 4) = 3;

      SmallMatrix<const double> X_exact = [] {
        SmallMatrix<double> Y{5, 3};
        Y(0, 0) = 1;
        Y(1, 0) = 3;
        Y(2, 0) = 2;
        Y(3, 0) = 4;
        Y(4, 0) = 5;

        Y(0, 1) = -3;
        Y(1, 1) = 6;
        Y(2, 1) = 1;
        Y(3, 1) = -2;
        Y(4, 1) = 4;

        Y(0, 2) = -2;
        Y(1, 2) = -4;
        Y(2, 2) = 2;
        Y(3, 2) = 6;
        Y(4, 2) = 7;

        return Y;
      }();

      SmallMatrix<double> b = A * X_exact;

      SmallMatrix<double> X{5, 3};

      Givens::solveCollection(A, X, b);
      SmallMatrix<double> error = X - X_exact;

      double max_error = 0;
      for (size_t i = 0; i < error.numberOfRows(); ++i) {
        for (size_t j = 0; j < error.numberOfColumns(); ++j) {
          max_error = std::max(max_error, std::abs(error(i, j)));
        }
      }
      REQUIRE(max_error < 1E-10);
    }

    SECTION("rectangular matrix")
    {
      SmallMatrix<double> A{5, 3};
      A.fill(0);
      A(0, 0) = 2;
      A(0, 1) = -1;

      A(1, 0) = -0.2;
      A(1, 1) = 2;
      A(1, 2) = -1;

      A(2, 1) = -1;
      A(2, 2) = 4;

      A(3, 2) = -1;
      A(4, 0) = 0.5;
      A(4, 1) = 1;
      A(4, 2) = 1;

      SmallMatrix<const double> X_exact = [] {
        SmallMatrix<double> Y{3, 4};
        Y(0, 0) = 1;
        Y(1, 0) = 3;
        Y(2, 0) = 2;

        Y(0, 1) = -1;
        Y(1, 1) = 1.5;
        Y(2, 1) = 5;

        Y(0, 2) = -3;
        Y(1, 2) = 1;
        Y(2, 2) = 4;

        Y(0, 3) = 0.7;
        Y(1, 3) = -3.2;
        Y(2, 3) = 2.5;

        return Y;
      }();

      SmallMatrix<double> B = A * X_exact;

      SmallMatrix<double> X{3, 4};
      X = zero;

      Givens::solveCollection(A, X, B);
      SmallMatrix<double> error = X - X_exact;

      double max_error = 0;
      for (size_t i = 0; i < error.numberOfRows(); ++i) {
        for (size_t j = 0; j < error.numberOfColumns(); ++j) {
          max_error = std::max(max_error, std::abs(error(i, j)));
        }
      }
      REQUIRE(max_error < 1E-10);
    }
  }
}
