#include <catch2/catch_approx.hpp>
#include <catch2/catch_test_macros.hpp>
#include <catch2/matchers/catch_matchers_all.hpp>

#include <mesh/NamedBoundaryDescriptor.hpp>
#include <mesh/NumberedBoundaryDescriptor.hpp>
#include <scheme/PolynomialReconstructionDescriptor.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("PolynomialReconstructionDescriptor", "[scheme]")
{
  SECTION("degree 1")
  {
    PolynomialReconstructionDescriptor descriptor{IntegrationMethodType::cell_center, 1};

    REQUIRE(descriptor.degree() == 1);
    REQUIRE(descriptor.stencilDescriptor().numberOfLayers() == 1);
    REQUIRE(descriptor.stencilDescriptor().connectionType() == StencilDescriptor::ConnectionType::by_nodes);
    REQUIRE(descriptor.symmetryBoundaryDescriptorList().size() == 0);

    REQUIRE(descriptor.preconditioning() == true);
    REQUIRE(descriptor.rowWeighting() == true);
  }

  SECTION("degree 4")
  {
    PolynomialReconstructionDescriptor descriptor{IntegrationMethodType::cell_center, 4};

    REQUIRE(descriptor.degree() == 4);
    REQUIRE(descriptor.stencilDescriptor().numberOfLayers() == 4);
    REQUIRE(descriptor.stencilDescriptor().connectionType() == StencilDescriptor::ConnectionType::by_nodes);
    REQUIRE(descriptor.symmetryBoundaryDescriptorList().size() == 0);

    REQUIRE(descriptor.preconditioning() == true);
    REQUIRE(descriptor.rowWeighting() == true);
  }

  SECTION("degree and stencil")
  {
    StencilDescriptor sd{2, StencilDescriptor::ConnectionType::by_faces};

    PolynomialReconstructionDescriptor descriptor{IntegrationMethodType::cell_center, 1, sd};

    REQUIRE(descriptor.degree() == 1);
    REQUIRE(descriptor.stencilDescriptor().numberOfLayers() == 2);
    REQUIRE(descriptor.stencilDescriptor().connectionType() == StencilDescriptor::ConnectionType::by_faces);
    REQUIRE(descriptor.symmetryBoundaryDescriptorList().size() == 0);

    REQUIRE(descriptor.preconditioning() == true);
    REQUIRE(descriptor.rowWeighting() == true);
  }

  SECTION("degree and symmetries")
  {
    std::vector<std::shared_ptr<const IBoundaryDescriptor>> bc_list;
    bc_list.push_back(std::make_shared<NamedBoundaryDescriptor>("XMIN"));
    bc_list.push_back(std::make_shared<NamedBoundaryDescriptor>("YMIN"));
    bc_list.push_back(std::make_shared<NumberedBoundaryDescriptor>(2));

    PolynomialReconstructionDescriptor descriptor{IntegrationMethodType::cell_center, 2, bc_list};

    REQUIRE(descriptor.degree() == 2);
    REQUIRE(descriptor.stencilDescriptor().numberOfLayers() == 2);
    REQUIRE(descriptor.stencilDescriptor().connectionType() == StencilDescriptor::ConnectionType::by_nodes);
    REQUIRE(descriptor.symmetryBoundaryDescriptorList().size() == 3);

    REQUIRE(descriptor.symmetryBoundaryDescriptorList()[0]->type() == IBoundaryDescriptor::Type::named);
    REQUIRE(descriptor.symmetryBoundaryDescriptorList()[1]->type() == IBoundaryDescriptor::Type::named);
    REQUIRE(descriptor.symmetryBoundaryDescriptorList()[2]->type() == IBoundaryDescriptor::Type::numbered);

    REQUIRE(descriptor.preconditioning() == true);
    REQUIRE(descriptor.rowWeighting() == true);
  }

  SECTION("degree, stencil and symmetries")
  {
    StencilDescriptor sd{3, StencilDescriptor::ConnectionType::by_edges};

    std::vector<std::shared_ptr<const IBoundaryDescriptor>> bc_list;
    bc_list.push_back(std::make_shared<NamedBoundaryDescriptor>("XMIN"));
    bc_list.push_back(std::make_shared<NumberedBoundaryDescriptor>(2));
    bc_list.push_back(std::make_shared<NamedBoundaryDescriptor>("YMIN"));

    PolynomialReconstructionDescriptor descriptor{IntegrationMethodType::cell_center, 1, sd, bc_list};

    REQUIRE(descriptor.degree() == 1);
    REQUIRE(descriptor.stencilDescriptor().numberOfLayers() == 3);
    REQUIRE(descriptor.stencilDescriptor().connectionType() == StencilDescriptor::ConnectionType::by_edges);
    REQUIRE(descriptor.symmetryBoundaryDescriptorList().size() == 3);

    REQUIRE(descriptor.symmetryBoundaryDescriptorList()[0]->type() == IBoundaryDescriptor::Type::named);
    REQUIRE(descriptor.symmetryBoundaryDescriptorList()[1]->type() == IBoundaryDescriptor::Type::numbered);
    REQUIRE(descriptor.symmetryBoundaryDescriptorList()[2]->type() == IBoundaryDescriptor::Type::named);

    REQUIRE(descriptor.preconditioning() == true);
    REQUIRE(descriptor.rowWeighting() == true);
  }

  SECTION("utlities")
  {
    PolynomialReconstructionDescriptor descriptor{IntegrationMethodType::cell_center, 3};

    REQUIRE(descriptor.degree() == 3);
    REQUIRE(descriptor.preconditioning() == true);
    REQUIRE(descriptor.rowWeighting() == true);

    SECTION("set preconditioning")
    {
      descriptor.setPreconditioning(false);

      REQUIRE(descriptor.degree() == 3);
      REQUIRE(descriptor.preconditioning() == false);
      REQUIRE(descriptor.rowWeighting() == true);

      descriptor.setPreconditioning(true);

      REQUIRE(descriptor.degree() == 3);
      REQUIRE(descriptor.preconditioning() == true);
      REQUIRE(descriptor.rowWeighting() == true);
    }

    SECTION("set row weighting")
    {
      descriptor.setRowWeighting(false);

      REQUIRE(descriptor.degree() == 3);
      REQUIRE(descriptor.preconditioning() == true);
      REQUIRE(descriptor.rowWeighting() == false);

      descriptor.setRowWeighting(true);

      REQUIRE(descriptor.degree() == 3);
      REQUIRE(descriptor.preconditioning() == true);
      REQUIRE(descriptor.rowWeighting() == true);
    }
  }
}
